//------------------------------------------------------------------------
//
// OpenDCC - DCC Sniffer
//
// Copyright (c) 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      tracker.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// webpage:   http://www.opendcc.de
// history:   2008-04-02 V0.01 kw started; based on the work of Alex Shephard
//                             see www.sourceforge.net, OpenDCC, DCCsniffer
//                             changes made: added individual trackers
//                                           added individual output flags
//                                           put strings to Flash
//            2008-04-17 V0.02 kw added LED control
//            2008-05-15 V0.03 kw added Trigger input
//            2008-05-26 V0.04 kw changed bitfield for FG2 and 3 to 4 Bits
//                                changed text F20-f12 to F20-F13
//                                Bugfix in Verbose Mode
//            2009-05-16 V0.05 kw added Analog Function
//            2009-06-23 V0.06 kw added DCC FAST CLOCK command
//                                changed call parameter to displayMultiFunctionMessage
//            2009-07-30          double IDLE removed
//            2010-03-17 V0.09 kw changed decoding of accessory address to native DCC
//            2010-08-12 V0.10 kw added register mode programming (3-4)
//
// to do:     accessory PoM commands
//
//------------------------------------------------------------------------
//
// purpose:   DCC sniffer
//            here: evaluate the DCC command and determine the
//            messages to be sent to the host.
//            internal track states are used 
//
// content:   1. States
//            2  Evaluate
//
// uses:      Interface Library (rprintf)
//            LED Conroller (led_control)
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <ctype.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <string.h>

#include "config.h"
#include "hardware.h"            // get LED defines
#include "led_control.h"         // control the local LEDs
#include "rprintf.h"             // small printf lib from pascal stang
#include "DccRx.h"               // structure defines of the DCC Message
#include "TimerPWM.h"            // measure_time
#include "external_trig.h"       // external trigger input

#include "tracker.h"

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine
                                 // 3: test action

//=============================================================================
// 1. Definitions
//=============================================================================
//
//
// #define LED_PROG     0     // program   - yellow
// #define LED_RAILCOM  1     // railcom   - yellow
// #define LED_ACC      2     // accessory - green
// #define LED_LOCO     3     // loco      - green
// #define LED_GO       4     // sniffing  - blue
// #define LED_ON       5     // on        - green


t_SnifferState SnifferState;

t_SnifferState eeConfig EEMEM =
  {
    (0<<0) |      // KeyMode: Bit 0: logg only during key pressed
    (1<<1) |      //          Bit 1: key toggles Enable State
    (1<<2),       //          Bit 2: reset timer

    5 |           // DisplayMode -> default Hex5
    D_HEX |       // Display Hex
    D_TSTAMP |    // Display Time Stamp
    D_VERBOSE,    // Display Verbose Messages

    D_ACTIVE |     // TrackTrigger;  
    D_INACTIVE,
    0xff,         // TrackIdle;     // show DCC Idle
    0xff,         // TrackSM;       // show ServiceMode
    0xff,         // TrackReset;    // show Resets
    0xff,         // TrackFastClk;  // show Clock
    0xff,         // TrackBiDi;     // show BiDi (Railcom)
    { 0x80,       // TrackLocoSpeed (up to 4 Trackers)
      {3,0,0,0}},    // Loco 
    { 0x80,       // TrackLocoFunc (up to 4 Trackers) 
      {3,0,0,0}},    // Functions
    { 0x80,       // TrackLocoPoM (up to 4 Trackers) 
      {3,0,0,0}},    // PoM
    { 0x80,       // TrackAcc (up to 4 Trackers)
      {5,6,7,8}},  
    { 0x80,       // TrackExt (up to 4 Trackers)
      {0,0,0,0}},    // ExtAccessory
  };


//------------------------------------------------------------------------------
// State Manager

void load_states(void)
  {
	unsigned char i;
    unsigned char  *ramptr;
    
    ramptr = (unsigned char *)&SnifferState;
    
    for (i=0; i < sizeof(SnifferState); i++)
      {
        *ramptr++ = eeprom_read_byte((unsigned char *)&eeConfig+i);
      }
  }	

void save_state_changes(void)
  {
	unsigned char i;
    unsigned char  *ramptr;
    
    ramptr = (unsigned char *)&SnifferState;
    
    for (i=0; i < sizeof(SnifferState); i++)
      {
    	if( *ramptr != eeprom_read_byte((unsigned char *)&eeConfig+i ) )
				eeprom_write_byte( (unsigned char *)&eeConfig+i, *ramptr ) ;
        ramptr++;
      }
  }	



//------------------------------------------------------------------------
// 2. Tracker
//------------------------------------------------------------------------
//

unsigned char should_we_track(t_trackpara *track, unsigned int addr)
  {
    if (SnifferState.DisplayMode & D_VERBOSE) 
      {
        if ( track->bits  & 0x80) return(TRUE);
        if ((track->bits  & (1<<0)) && (track->addr[0] == addr)) return(TRUE);
        if ((track->bits  & (1<<1)) && (track->addr[1] == addr)) return(TRUE);
        if ((track->bits  & (1<<2)) && (track->addr[2] == addr)) return(TRUE);
        if ((track->bits  & (1<<3)) && (track->addr[3] == addr)) return(TRUE);
      }
    return(FALSE);
  }

//------------------------------------------------------------------------
// 2.a) Strings in ProgMemory

char text_stop[]           PROGMEM = "Stop";
char text_estop[]          PROGMEM = "E-Stop";
char text_reserved[]       PROGMEM = "Reserved";
char text_bin_state_ctrl[] PROGMEM = "Bin State Ctrl: [";
char text_clock[]          PROGMEM = "Clk: ";   // Clk: 10:45;su*5
char text_dec_ack[]        PROGMEM = "Decoder Acknowledgment";
char text_hard_reset[]     PROGMEM = "Hard Reset";
char text_factory_test[]   PROGMEM = "Factory Test";
char text_dec_flags[]      PROGMEM = "Set Decoder Flags";
char text_adv_addr[]       PROGMEM = "Set Advanced Addressing";
char text_feature_ex[]     PROGMEM = "Feature Expansion, ";
char text_idle[]           PROGMEM = "Idle";
char text_bd[]             PROGMEM = "BiDi ";
char text_anlg[]           PROGMEM = "Anlg.Fct ";
char text_cv_access[]      PROGMEM = "CV Access: ";
char text_2dash[]          PROGMEM = "--";



//------------------------------------------------------------------------------

// 2.b) Decoderroutines
//------------------------------------------------------------------------
uint8_t displayTimeStamp(void)
  {
    if (SnifferState.DisplayMode & D_TSTAMP) 
      {
        rprintfNum( 10, 4, 0, ' ', measure_time / 1000) ;
        rprintfChar( ',' );
        rprintfNum( 10, 3, 0, '0', measure_time % 1000) ;
        rprintfChar( 's' );
        rprintfChar( ' ' );
        return(TRUE);
      }
    return(FALSE);
  }

uint8_t displayDccMessageHex(DCC_MSG * pDccMsg)
  {
    uint8_t i;

    if( SnifferState.DisplayMode & D_HEX )
      {
        // Print a leading 'P'  or 'R'
        if ( SnifferState.DisplayMode & 0x04 )
          {
            if ( pDccMsg->Qualifier & MSG_RAILCOM ) rprintfChar( 'R' ) ;
            else rprintfChar( 'P' ) ;
          }

        // Output the Preamble bit count
        if( ( SnifferState.DisplayMode & 0x02 ) == 0 )
          {
            if( SnifferState.DisplayMode  <= 1 )
                rprintfChar( pDccMsg->PreambleBits + 22 ) ;
            else
                rprintfu08( pDccMsg->PreambleBits ) ;
          }

        for(i = 0; i < pDccMsg->Size; i++ )
          {
            // Print a leading <space> except for the first byte of mode 3  
            if( ( SnifferState.DisplayMode  & 0x01 ) && ( ( SnifferState.DisplayMode  != 3 ) || i ) )  
                rprintfChar( ' ' ) ;

            rprintfu08( pDccMsg->Data[i]) ;
          }
        rprintfChar( ' ' ) ;
        return(TRUE);
      }
    return(FALSE);
  }

uint8_t displayTimeAndHex(DCC_MSG * pDccMsg)
  {
    uint8_t Actioned;

    Actioned = displayTimeStamp();
    Actioned |= displayDccMessageHex( pDccMsg );
    return(Actioned);
  }

uint8_t displayProgMessage( uint8_t Data0, uint8_t Data1, uint8_t Data2 )
  {
    unsigned cv_addr;

    cv_addr = ((Data0 & 0x03) << 8) + Data1 + 1;

    if (SnifferState.DisplayMode & D_VERBOSE)
      {
        if ((Data0 & 0x0C) == 0x00)
          {
            rprintfProgStr( text_reserved );     //rprintfStr( "Reserved" );
          }
        else if ((Data0 & 0x0C) == 0x04)
          {
             //  preamble 0 0111 CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1 mit CC=01  = CV-Mode, verify
            rprintfStr( "Verify: " ) ;
            rprintfNum( 10, 4, 0, '0', cv_addr ) ;
            rprintfChar( ':' );
            rprintfNum( 10, 4, 0, '0', Data2 ) ;
          }
        else if ((Data0 & 0x0C) == 0x08)
          {
            //  preamble 0 0111 CCAA 0 AAAAAAAA 0 111KDBBB 0 EEEEEEEE 1 mit CC=10  = CV-Mode, single Bit
            //  K = (1=write, 0=verify) D = Bitvalue, BBB = bitpos
            rprintfStr( "Bit-" ) ;
            if (Data2 & 0x10) rprintfStr( "Wr:" ) ;
            else  rprintfStr( "Vf:" ) ;
            rprintfNum( 10, 4, 0, '0', cv_addr ) ;
            rprintfChar( ',' );
            rprintfNum( 10, 1, 0, '0', Data2 & 0x07 ) ;
            rprintfChar( ':' );
            if (Data2 & 0x08) rprintfChar( '1' );
            else  rprintfChar( '0' );
          }
        else if ((Data0 & 0x0C) == 0x0C)
          {
            //  preamble 0 0111 CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1 mit CC=11  = CV-Mode, write
            rprintfStr( "Write: " ) ;
            rprintfNum( 10, 4, 0, '0', cv_addr ) ;
            rprintfChar( ':' );
            rprintfNum( 10, 4, 0, '0', Data2 ) ;
          }
        return(TRUE);
      }
    return(FALSE);
  }


uint8_t displayProgRegMessage( uint8_t Data0, uint8_t Data1)
  {
    // long-preamble 0 0111CRRR 0 DDDDDDDD 0 EEEEEEEE 1
    // RRR=Register address (range 0...7)
    // C=0 Verify, C=1 Write register
    // RRR=000 ist wie Adress only!

    if (SnifferState.DisplayMode & D_VERBOSE)
      {
        if ((Data0 & 0x08) == 0x00)
          {
             //  preamble 0 0111CRRR 0 DDDDDDDD 0 EEEEEEEE 1 mit C=0 Verify
            rprintfStr( "RegVfy: " ) ;
          }
        else
          {
             //  preamble 0 0111CRRR 0 DDDDDDDD 0 EEEEEEEE 1 mit C=1 Write register
            rprintfStr( "RegWrt: " ) ;
          }
        rprintfNum( 10, 1, 0, '0', (Data0 & 0x07) + 1  ) ;
        rprintfChar( ':' );
        rprintfNum( 10, 4, 0, '0', Data1 ) ;
        return(TRUE);
      }
    return(FALSE);
  }



uint8_t displayLocoAddress(DCC_MSG * pDccMsg)
  {
    uint8_t Actioned;

    Actioned = displayTimeAndHex( pDccMsg );

    if (SnifferState.DisplayMode & D_VERBOSE)
      {
        rprintfChar( 'L' ) ;
        if (pDccMsg->Data[0] < 112)
          {
            rprintfChar( '.' ) ;                // short address padded with points
            rprintfChar( '.' ) ;
            rprintfNum( 10, 3, 0, '0', pDccMsg->Data[0]) ;
          }
        else
          {
            uint16_t Address;
            Address = ( (pDccMsg->Data[0] & 0x3F) << 8 ) | pDccMsg->Data[1];
            rprintfNum( 10, 5, 0, '0', Address) ;  // long address
          }
        rprintfChar(' ');
        Actioned = 1 ;
      }
    return(Actioned);
  }

// Parameters: pDccMsg: pointer to complete message
//             Cmd:     index of command byte (the byte after the address)

uint8_t displayMultiFunctionMessage(DCC_MSG * pDccMsg, uint8_t c_index)
  {
    uint8_t Actioned = 0;

    uint8_t Cmd = pDccMsg->Data[c_index];
    uint8_t Data1 = pDccMsg->Data[c_index+1];
    uint8_t Data2 = pDccMsg->Data[c_index+2];

    uint16_t Address;
    if (pDccMsg->Data[0] < 112)
        Address = pDccMsg->Data[0];
    else
        Address = ( (pDccMsg->Data[0] & 0x3F) << 8 ) | pDccMsg->Data[1];

    // 3 MSBs determine command type
    switch( Cmd & 0xE0 )
      {
        case 0x00:  // Decoder and Consist Control
            if (should_we_track (&SnifferState.TrackLocoPoM, Address) )
              {
                Actioned = displayLocoAddress(pDccMsg);
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    switch( Cmd & 0x0E )
                      {
                        case 0x00:
                            if( Cmd & 0x01 ) // We only display the Hard Reset here as the other one is done at the outer level 
                            rprintfProgStr( text_hard_reset );   //rprintfStr( "Hard Reset" );
                            break ;
                        case 0x02:
                            rprintfProgStr( text_factory_test ); //rprintfStr( "Factory Test" );
                            break ;
                        case 0x06:
                            rprintfProgStr( text_dec_flags );    //rprintfStr( "Set Decoder Flags" );
                            break ;
                        case 0x0A:
                            rprintfProgStr( text_adv_addr );     //rprintfStr( "Set Advanced Addressing" );
                            break ;
                        case 0x0E:
                            rprintfProgStr( text_dec_ack );      //rprintfStr( "Decoder Acknowledgment" );
                            break ;
                        default:
                            rprintfProgStr( text_reserved );     //rprintfStr( "Reserved" );
                      }
                    Actioned = 1 ;
                  }
                
              }
            break ;

        case 0x20:  // Advanced Operations (128 Speed Steps)
            if (should_we_track (&SnifferState.TrackLocoSpeed, Address) )
              {
                Actioned = displayLocoAddress(pDccMsg);
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    switch( Cmd & 0x1F )
                      {
                        case 0x1F:
                            if (SnifferState.DisplayMode & D_VUMETER)
                              {
                                rprintfChar( 'S' );
                                switch( Data1 & 0x7F )
                                  {
                                    case 0x00:  
                                        rprintfChar( '0' ); 
                                        break ;
                                    case 0x01:  
                                        rprintfChar( 'X' ); //rprintfStr( "E-Stop" ) ;
                                        break ;
                                    default:   // Speed
                                        {
                                        uint8_t si, speed, dir;
                                        speed = Data1 & 0x7F;
                                        speed = speed / 4;  // scale down to 0..31
                                        dir = (Data1 & 0x80) ? '+' : '-';
                                        for (si = 0; si < speed; si++)
                                          {
                                            rprintfChar( dir ); 
                                          }
                                        }   
                                  }
                              }
                            else
                              {
                                switch( Data1 & 0x7F )
                                  {
                                    case 0x00:  
                                        rprintfProgStr( text_stop );  //rprintfStr( "Stop" ) ;
                                        break ;
                                    case 0x01:  
                                        rprintfProgStr( text_estop );  //rprintfStr( "E-Stop" ) ;
                                        break ;
                                    default:   // Speed
                                        rprintfChar( 'S' );
                                        rprintfNum( 10, 3, 0, '0', (Data1 & 0x7F) - 1 ) ;
                                        rprintfChar( (Data1 & 0x80) ? 'F' : 'R' ) ;
                                  }
                              }
                            break;
                        case 0x1D:
                            rprintfProgStr( text_anlg );  //rprintfStr( "Anlg.Fct " ) ;
                            rprintfNum( 10, 3, 0, '0', Data1 ) ;
                            rprintfChar( ':' ) ;
                            rprintfNum( 10, 3, 0, '0', Data2 ) ;
                            break;
                        default:
                            break;
                      }
                    Actioned = 1;
                  }
              }
            break;
        case 0x40:  // Speed and Direction Instruction for reverse operation
        case 0x60:  // 011 Speed and Direction Instruction for forward operation
            // 01DUSSSS
            if (should_we_track (&SnifferState.TrackLocoSpeed, Address) )
              {
                Actioned = displayLocoAddress(pDccMsg);
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    if (SnifferState.DisplayMode & D_VUMETER)
                      {
                        rprintfChar( 'S' );
                        switch( Cmd & 0x1F )
                          {
                            case 0x00:  
                            case 0x10:  
                                rprintfChar( '0' ); 
                                break ;
                            case 0x01:  
                            case 0x11:  
                                rprintfChar( 'X' ); //rprintfStr( "E-Stop" ) ;
                                break ;
                            default:   // Speed
                                {
                                uint8_t si, speed, dir;
                                speed = (Cmd & 0x0F) * 2 + (Cmd & 0x10 >> 4);
                                speed = speed - 3;  // scale down to 0..31
                                dir = (Cmd & 0x20) ? '+' : '-';
                                for (si = 0; si < speed; si++)
                                  {
                                    rprintfChar( dir ); 
                                  }
                                }   
                          }
                      }
                    else
                      {
                        switch( Cmd & 0x1F )
                          {
                            case 0x00:  
                            case 0x10:
                                rprintfProgStr( text_stop );   //rprintfStr( "Stop" ) ;
                                break ;
                            case 0x01:  
                            case 0x11:
                                rprintfProgStr( text_estop );  //rprintfStr( "E-Stop" ) ;
                                break ;
                            default:
                                rprintfChar( 'S' ) ;
                                rprintfNum( 10, 2, 0, '0', (((Cmd & 0x0F) << 1 ) | ((Cmd & 0x10) >> 4)) - 3 ) ;
                                rprintfChar( (Cmd & 0x20) ? 'F' : 'R' ) ;
                          }
                        Actioned = 1 ;
                      }
                  }
              }
            break;
    
        case 0x80:  // 100FFFFF
            if (should_we_track (&SnifferState.TrackLocoFunc, Address) )
              {
                Actioned = displayLocoAddress(pDccMsg);
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    rprintfStr( "FG1: " ) ;
                    rprintfNum( 2, 5, 0, '0', Cmd & 0x1F ) ;
                    Actioned = 1 ;
                  }
              }
            break;
    
        case 0xA0:  // 1010FFFF  // 1010FFFF
            if (should_we_track (&SnifferState.TrackLocoFunc, Address) )
              {
                Actioned = displayLocoAddress(pDccMsg);
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    rprintfStr( "FG2: " ) ;
                    if (Cmd & 0x10) 
                      {
                        rprintfStr( "F8-F5: " ) ;
                        rprintfNum( 2, 4, 0, '0', Cmd & 0x0F ) ;
                      }
                    else 
                      {
                        rprintfStr( "F12-F9: " ) ;
                        rprintfNum( 2, 4, 0, '0', Cmd & 0x0F ) ;
                      }
                    Actioned = 1 ;
                  }
              }
            break;

        case 0xC0:  // 110CCCCC
            switch( Cmd & 0x1F )
              {
                case 0x1F:   // F28-F21
                    if (should_we_track (&SnifferState.TrackLocoFunc, Address) )
                      {
                        Actioned = displayLocoAddress(pDccMsg);
                        if (SnifferState.DisplayMode & D_VERBOSE)
                          {
                            rprintfStr( "F28-F21: " ) ;
                            rprintfNum( 2, 8, 0, '0', Data1 ) ;
                            Actioned = 1 ;
                          }
                      }
                    break;
                case 0x1E:   // F20-F13
                    if (should_we_track (&SnifferState.TrackLocoFunc, Address) )
                      {
                        Actioned = displayLocoAddress(pDccMsg);
                        if (SnifferState.DisplayMode & D_VERBOSE)
                          {
                            rprintfStr( "F20-F13: " ) ;
                            rprintfNum( 2, 8, 0, '0', Data1 ) ;
                            Actioned = 1 ;
                          }
                      }
                    break;
                case 0x1D: // CCCCC = 11101:  Binary State Control Instruction short form
                           // following: DLLLLLLL
                    if (should_we_track (&SnifferState.TrackLocoPoM, Address) )
                      {
                        Actioned = displayLocoAddress(pDccMsg);
                        if (SnifferState.DisplayMode & D_VERBOSE)
                          {
                            rprintfProgStr( text_bin_state_ctrl );// rprintfStr( "Bin State Ctrl: [" ) ;
                            rprintfNum( 10, 5, 0, '.', (Data1 & 0x7F)  ) ;
                            rprintfStr( "]=" ) ;
                            rprintfChar( (Data1 & 0x80) ? '1' : '0' ) ;                
                            Actioned = 1 ;
                          }
                      }
                    break;
                case 0x00:  // CCCCC = 00000 = binary states control (long form)
                            // following: DLLLLLLL HHHHHHHH.
                            // means D at HHHHHHHHLLLLLLL (0=broadcast all states).
                    if (should_we_track (&SnifferState.TrackLocoPoM, Address) )
                      {
                        Actioned = displayLocoAddress(pDccMsg);
                        if (SnifferState.DisplayMode & D_VERBOSE)
                          {
                            rprintfProgStr( text_bin_state_ctrl );// rprintfStr( "Bin State Ctrl: [" ) ;
                            rprintfNum( 10, 5, 0, '0', ((Data2<<7) | (Data1 & 0x7F))  ) ;
                            rprintfStr( "]=" ) ;
                            rprintfChar( (Data1 & 0x80) ? '1' : '0' ) ;                
                            Actioned = 1 ;
                          }
                      }
                    break;
                case 0x01:  // CCCCC = 00001 = fast clock 
                            // following: TCODE1, TCODE2, TCODE3, TCODE4.
                            // TCODEx = CCDDDDDD
                            // CC denotes the time code type, DDDDDD the corresponding data content.
                            // CC = 00: DDDDDD = mmmmmm, this denotes the minute, range 0..59.
                            // CC = 10: DDDDDD = 0HHHHHH, this denotes the hour, range 0..23
                            // CC = 01: DDDDDD = 000WWW, this denotes the day of week, 
                            //          0=Monday, 1=Tuesday, 2=Wednesday, 3=Thursday, 4=Friday, 5=Saturday, 6=Sunday.
                            // CC = 11: DDDDDD = 00FFFFF, this denotes the acceleration factor,
                            
                    if( SnifferState.TrackFastClk )
                      {
                        Actioned = displayLocoAddress(pDccMsg);
                        if (SnifferState.DisplayMode & D_VERBOSE)
                          {
                            rprintfProgStr( text_clock );// rprintfStr( "Clk: " ) ; // Clk: 10:45;su*5
                            // now check all tcodes

                            uint8_t minute = 99;
                            uint8_t hour = 99;
                            uint8_t wday = 99;
                            uint8_t factor = 99;

                            uint8_t j;
                            for(j = c_index+1; j < pDccMsg->Size - 1 ; j++ )
                              {
                                switch(pDccMsg->Data[j] & 0xC0)
                                  {
                                    case 0x00: minute = pDccMsg->Data[j] & 0x3F;
                                        break;
                                    case 0x40: wday = pDccMsg->Data[j] & 0x3F;
                                        break;
                                    case 0x80: hour = pDccMsg->Data[j] & 0x3F;
                                        break;
                                    case 0xC0: factor = pDccMsg->Data[j] & 0x3F;
                                        break;
                                  }
                              }                            
                            if (hour != 99) rprintfNum( 10, 2, 0, '0', hour ) ;
                            else rprintfProgStr( text_2dash );
                            rprintfChar( ':' );
                            if (minute != 99) rprintfNum( 10, 2, 0, '0', minute ) ;
                            else rprintfProgStr( text_2dash );
                            rprintfChar( ';' );
                            switch(wday)
                              {
                                default: rprintfProgStr( text_2dash );  break;
                                case 0:  rprintfChar( 'm' ); rprintfChar( 'o' ); break;
                                case 1:  rprintfChar( 't' ); rprintfChar( 'u' ); break;
                                case 2:  rprintfChar( 'w' ); rprintfChar( 'e' ); break;
                                case 3:  rprintfChar( 't' ); rprintfChar( 'h' ); break;
                                case 4:  rprintfChar( 'f' ); rprintfChar( 'r' ); break;
                                case 5:  rprintfChar( 's' ); rprintfChar( 'a' ); break;
                                case 6:  rprintfChar( 's' ); rprintfChar( 'u' ); break;
                              }
                            rprintfChar( '@' );
                            if (factor != 99) rprintfNum( 10, 2, 0, '0', factor ) ;
                            else rprintfProgStr( text_2dash );
                            Actioned = 1 ;
                          }
                      }
                    break;
                default:
                    if (should_we_track (&SnifferState.TrackLocoPoM, Address) )
                      {
                        Actioned = displayLocoAddress(pDccMsg);
                        if (SnifferState.DisplayMode & D_VERBOSE)
                          {
                            rprintfProgStr( text_feature_ex ); //rprintfStr( "Feature Expansion, " ) ;
                            rprintfProgStr( text_reserved );  //rprintfStr( "Reserved" );
                            Actioned = 1 ;
                          }
                      }
              } 
            break;

        case 0xE0:
            if (should_we_track (&SnifferState.TrackLocoPoM, Address) )
              {          
                Actioned = displayLocoAddress(pDccMsg);
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    rprintfProgStr( text_cv_access ); //  rprintfStr( "CV Access: " ) ;
                    displayProgMessage(Cmd, Data1, Data2);  // here we do not look for register mode!
                    Actioned = 1 ;
                  }
              }
            break;
      }
    return(Actioned);
  }


/*
unsigned char should_we_track_loco (uint16_t Address)
  {
    if( should_we_track (&SnifferState.TrackLocoSpeed, Address ) ) return(TRUE);
    if( should_we_track (&SnifferState.TrackLocoFunc, Address ) ) return(TRUE);
    if( should_we_track (&SnifferState.TrackLocoPoM, Address ) ) return(TRUE);
    return(FALSE);
  }
*/

void displayDccMessage( DCC_MSG	*pDccMsg )
  {
    uint16_t Address ;
    
    uint8_t Actioned = 0 ;
  
    if ( ( pDccMsg->Data[0] == 0 ) && ( pDccMsg->Data[1] == 0 ) )
      {  
        if( SnifferState.TrackReset )
          {
            Actioned = displayTimeAndHex( pDccMsg );        
            if (SnifferState.DisplayMode & D_VERBOSE)
              {
                rprintfStr( "Reset" );
                Actioned = 1 ;
              } 
          }
      }
  
    // Multi Function Decoders (7-bit address)
    else if( ( pDccMsg->Data[0] < 112 ) )
      {
        // Address = pDccMsg->Data[0];
        // command starts at Data1
        if( displayMultiFunctionMessage(pDccMsg, 1 ) )  
          {
            led_highlight(LED_LOCO);
            Actioned = 1 ;
          }
        else
          {
            led_half_pulse(LED_LOCO);
          }
      }

    // Service Mode Instructions
    else if( ( pDccMsg->Data[0] < 128 ) )
      {
        if( SnifferState.TrackSM )
          {
            led_highlight(LED_PROG);
            Actioned = displayTimeAndHex( pDccMsg );
            if (SnifferState.DisplayMode & D_VERBOSE)
              {
                rprintfStr( "Program: " ) ;
                if (pDccMsg->Size == 4)
                    displayProgMessage (pDccMsg->Data[0],pDccMsg->Data[1],pDccMsg->Data[2]);
                else if (pDccMsg->Size == 3)
                    displayProgRegMessage (pDccMsg->Data[0],pDccMsg->Data[1]);
              }
            Actioned = 1 ;
          }
        else
          {
            led_half_pulse(LED_LOCO);
          }
      }
      
    // Basic Accessory Decoders (9-bit) & Extended Accessory Decoders (11-bit)
    else if( pDccMsg->Data[0] < 192 )
      {
        // Address according to IB-Scheme
        Address = ((((~pDccMsg->Data[1]) & 0x70) << 2 ) | (pDccMsg->Data[0] & 0x3F) ) - 1 ;
        Address = ( Address << 2 ) | ( ( (pDccMsg->Data[1] & 0x06) >> 1 ) + 1 ) ;

        // True DCC Address
        Address = ((((~pDccMsg->Data[1]) & 0x70) << 2 ) | (pDccMsg->Data[0] & 0x3F) );
        Address = ( Address << 2 ) | ( ( (pDccMsg->Data[1] & 0x06) >> 1 )     ) ;


        // this is either the full turnout address or the decoder address of an extended

        if (pDccMsg->Data[1] & 0x80)
          {  // basic
            if( should_we_track (&SnifferState.TrackAcc, Address) )
              {
                led_highlight(LED_ACC);
                Actioned = displayTimeAndHex( pDccMsg );
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    rprintfChar( 'A') ;
                    rprintfNum( 10, 4, 0, '0', Address ) ;
      
                    rprintfChar( (pDccMsg->Data[1] & 0x01) ? 'N' : 'R' ) ; 
                    rprintfChar( (pDccMsg->Data[1] & 0x08) ? '+' : '-' ) ; 
                    Actioned = 1 ;
                  }                
              }
            else
              {
                led_half_pulse(LED_ACC);
              }
          }
        else
          {  // extended
            if (pDccMsg->Data[2] & 0x08)
              { // Railcom NOP Command
                if( SnifferState.TrackBiDi )
                  {
                    Actioned = displayTimeAndHex( pDccMsg );
                    if (SnifferState.DisplayMode & D_VERBOSE)
                      {
                        rprintfProgStr( text_bd );
                        rprintfStr( "NOP:" );
                        rprintfNum( 10, 4, 0, '0', Address ) ;
                        Actioned = 1 ;
                      }
                  }
              }
            else
              {
                if( should_we_track (&SnifferState.TrackExt, Address) )
                  {
                    led_fast_pulse(LED_ACC);
                    Actioned = displayTimeAndHex( pDccMsg );
                    if (SnifferState.DisplayMode & D_VERBOSE)
                      {
                        rprintfChar( 'S' ) ;
                        rprintfNum( 10, 4, 0, '0', Address ) ;
                        rprintfChar( ':' ) ;
                        rprintfu08( pDccMsg->Data[2] ) ;
                        Actioned = 1 ;
                      }
                  }
                else
                  {
                    led_half_pulse(LED_ACC);
                  }
              }
          }
      }
  
    // Multi Function Decoders (14-bit address)
    else if( pDccMsg->Data[0] < 232 )
      {
        // Address = ( (pDccMsg->Data[0] & 0x3F) << 8 ) | pDccMsg->Data[1];
        // command starts at Data[2]
        if( displayMultiFunctionMessage(pDccMsg, 2) )
          {
            led_highlight(LED_LOCO);
            Actioned =  1 ;
          }
        else
          {
            led_half_pulse(LED_LOCO);
          }
      }

    
    // Reserved
    else if( pDccMsg->Data[0] < 255 ) 
      {
        Actioned = displayTimeAndHex( pDccMsg );

        if (SnifferState.DisplayMode & D_VERBOSE)
          {
            rprintfProgStr( text_reserved );  //rprintfStr( "Reserved" );
            Actioned = 1 ;
          }
      }
    // Idle
    else 
      {
        if( ( pDccMsg->Data[1] == 0 ) )
          {
            if( SnifferState.TrackIdle )
              {
                Actioned = displayTimeAndHex( pDccMsg );
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    rprintfProgStr( text_idle );
                    Actioned = 1 ;
                  } 
              }
          }
        else
          {
            if( SnifferState.TrackBiDi )
              {
                Actioned = displayTimeAndHex( pDccMsg );
                if (SnifferState.DisplayMode & D_VERBOSE)
                  {
                    unsigned char j;
                    rprintfProgStr( text_bd );
                    switch(pDccMsg->Data[1])
                      {
                        case 0x7F:              // Command Station ID for accessory
                            rprintfStr( "CS-IDa:" );
                            break;
                        case 0xFF:              // Command Station ID for mobile decoders
                            rprintfStr( "CS-IDm:" );
                            break;
                        case 0xFE:              // ID Search
                            rprintfStr( "Scan:" );
                            break;
                        case 0xFD:              // Direct Access
                            rprintfStr( "Direct:" );
                            break;
                        case 0xFC:              // write Address
                            rprintfStr( "WrAddr:" );
                            break;
                        default:
                            rprintfStr( "???:" );
                            break;
                      }
                    rprintfu08(pDccMsg->Data[2]);
                    rprintfChar( '-' ) ;
                    for(j = 3; j < pDccMsg->Size; j++ )
                      {
                        rprintfu08( pDccMsg->Data[j]) ;
                        if (j==7) rprintfChar( ':' );
                        else rprintfChar( ' ' );
                      }
                    Actioned = 1 ;
                  } 
              }
          }
      }

    if (Actioned)    rprintfCRLF();
  }



//-----------------------------------------------------------------------
// external Trigger
//-----------------------------------------------------------------------
// watch the trigger input and create a message

void watch_external_trigger(void)
  {
    unsigned char state;

    state = trigger_poll();

    if (state == 0)
      {
        if (SnifferState.TrackTrigger & D_INACTIVE)
          {
            displayTimeStamp();
            rprintfStr( "Trig inactive" );
            rprintfCRLF();        
          }
      }
    else if (state == 1)
      {
        if (SnifferState.TrackTrigger & D_ACTIVE)
          {
            displayTimeStamp();
            rprintfStr( "Trig active" );
            rprintfCRLF();        
          }
      }
  }


//-----------------------------------------------------------------------
// Railcom and polarity
//-----------------------------------------------------------------------

// run this as thread, to get blinking off again.

uint8_t railcom_Qualifier;

void show_railcom_polarity(uint8_t Qualifier)
  {
    if (Qualifier != railcom_Qualifier)
      {
        railcom_Qualifier = Qualifier;
        switch(railcom_Qualifier & (MSG_RAILCOM | MSG_POLARITY))
          {            
            case 0:
                led_XoooooXooooo(LED_RAILCOM); 
                break;
            case MSG_POLARITY:
                led_oXXXXXoXXXXX(LED_RAILCOM);
                break;
            case MSG_RAILCOM:
                led_XoooXooo(LED_RAILCOM);
                break;
            case MSG_RAILCOM | MSG_POLARITY:
                led_oXXXoXXX(LED_RAILCOM);
                break;
          }
      }
  }



//------------------------------------------------------------------------
// 
// 3. Upstream Interface
//
//------------------------------------------------------------------------

unsigned char EnableDisplay;

void toggle_EnableDisplay(void)
  {
    if (EnableDisplay)
      {
        EnableDisplay = FALSE;
        led_off(LED_GO);
      }
    else
      {
        EnableDisplay = TRUE;
        led_half(LED_GO);  
      }
  }


DCC_MSG	DccMsg;

int8_t last_msg;    // signed

void init_tracker(void)
  {
    load_states();              // get all states from eeprom
    EnableDisplay = FALSE;
    toggle_EnableDisplay();
    railcom_Qualifier = 0xff;   // set void
    last_msg = systick_1ms;     // keep track of last dcc msg
  }


char text_checksum_error[]   PROGMEM = "Checksum Error!";

void run_tracker(void)
  {
    watch_external_trigger();

    if( getDccRxMsg(&DccMsg) )
      {
        uint8_t i ;
        uint8_t xor = 0 ;

        for (i=0; i< DccMsg.Size; i++)
            xor = xor ^ DccMsg.Data[i];

        // if ( DccMsg.Qualifier & MSG_POLARITY ) led_highlight(LED_RAILCOM);  // show polarity

        last_msg = systick_1ms;
        show_railcom_polarity(DccMsg.Qualifier);

        if ( EnableDisplay)
          {
            if( xor )
              {
                displayTimeAndHex( &DccMsg );
                rprintfProgStr( text_checksum_error ) ;    // regardless of tracker states!
                rprintfCRLF() ;
              }
            else
                displayDccMessage( &DccMsg ) ;
          }
      }
      if ((char)(systick_1ms - last_msg) > 50)
      {
        // obviously timeout in DCC
        show_railcom_polarity(0xff);  // kill any internal state
        led_off(LED_RAILCOM);
        led_off(LED_LOCO);
        led_off(LED_ACC);
        led_off(LED_PROG);
      }
  }



